
from telegram import (ReplyKeyboardRemove, Update)
from telegram.ext import (CallbackContext)

from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate

def telegram_append_to_context(update: Update, context: CallbackContext, key, value):
    if not key in context.user_data:
        context.user_data[key] = []
    context.user_data[key].append(value)


def whatsapp_append_to_context(client, update: WhatsAppUpdate, context, key, value):
    if not key in context:
        context[key] = []
    context[key].append(value)


def telegram_check_in_context(update: Update, context: CallbackContext, key, value, doppelte_antwort):
    if key in context.user_data:
        if value in context.user_data[key]:
            update.message.reply_text(doppelte_antwort,
                                      reply_markup=ReplyKeyboardRemove())
            return "{}_FRAGE".format(key.upper())


def whatsapp_check_in_context(client, update: WhatsAppUpdate, context,  key, value, doppelte_antwort):
    if key in context:
        if value in context[key]:
            client.messages.create(
                body=doppelte_antwort,
                from_=update.To,
                to=update.From
            )
            return "{}_FRAGE".format(key.upper())


def telegram_eval_list(update: Update, context: CallbackContext, answer_id_name_list, poi, response_text):
    if not poi in context.user_data:
        context.user_data[poi] = []

    response_text += "\n"

    for id, name in answer_id_name_list:
        if id in context.user_data[poi]:
            response_text += "✅ {}\n".format(name)
        else:
            response_text += "◽ {}\n".format(name)

    update.message.reply_text(response_text,
                              reply_markup=ReplyKeyboardRemove())


def whatsapp_eval_list(client, update: WhatsAppUpdate, context, answer_id_name_list, poi, response_text):
    if not poi in context:
        context[poi] = []

    response_text += "\n"

    for id, name in answer_id_name_list:
        if id in context[poi]:
            response_text += "✅ {}\n".format(name)
        else:
            response_text += "◽ {}\n".format(name)

    client.messages.create(
        body=response_text,
        from_=update.To,
        to=update.From
    )


telegram_action_functions = {"append_to_context": telegram_append_to_context,
                             "check_in_context": telegram_check_in_context,
                             "eval_list": telegram_eval_list
                             }

whatsapp_action_functions = {"append_to_context": whatsapp_append_to_context,
                             "check_in_context": whatsapp_check_in_context,
                             "eval_list": whatsapp_eval_list
                             }



import mongoengine
from telegram.constants import MAX_MESSAGE_LENGTH
import datetime

from configparser import ConfigParser

config = ConfigParser()
config.read('config.ini')

dbname = config["bot"]["bot_name"]
class User(mongoengine.Document):
    user_id = mongoengine.StringField(required=True)
    first_name = mongoengine.StringField(required=True, max_length=50)
    last_name = mongoengine.StringField(max_length=50)
    username = mongoengine.StringField(max_length=50)
    language_code = mongoengine.StringField(max_length=10)
    entry_time = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}

class Interaction(mongoengine.Document):
    user = mongoengine.ReferenceField(User)
    state=mongoengine.StringField(max_length=50)
    update_id = mongoengine.IntField()
    first_name = mongoengine.StringField(max_length=50)
    last_name = mongoengine.StringField(max_length=50)
    username = mongoengine.StringField(max_length=50)
    interaction_full = mongoengine.DictField()
    interaction_text = mongoengine.StringField(max_length=MAX_MESSAGE_LENGTH)
    interaction_data = mongoengine.StringField(max_length=MAX_MESSAGE_LENGTH)
    interaction_id = mongoengine.IntField()
    date = mongoengine.DateTimeField(default=datetime.datetime.utcnow)
    meta = {'db_alias': dbname}





import re
EMOJI_PATTERN = re.compile(
    "["
    "\U0001F1E0-\U0001F1FF"  # flags (iOS)
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F700-\U0001F77F"  # alchemical symbols
    "\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
    "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
    "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    "\U0001FA00-\U0001FA6F"  # Chess Symbols
    "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027B0"  # Dingbats
    "\U000024C2-\U0001F251"
    "]+"
)

WEITER_PATTERN = re.compile(
    "^("
    "gefunden|"
    "bescheid|"
    "Weiter|"
    "Bescheid|"
    "Gefunden|"
    "weiter|"
    "next|"
    "nächster|"
    "weit|"
    "witer|"
    "weitr|"
    "überspringen|"
    "uberspringen|"
    "ueberspringen|"
    "ueber springen|"
    "waiter|"
    "bin soweit|"
    ")$"
)

WOHIN_PATTERN = re.compile(
    "^("
    "wohin|"
    "wo|"
    "weg|"
    "wo lang|"
    "route|"
    "ziel|"
    "ort|"
    "woin|"
    "treffpunkt|"
    "GPS|"
    "way|"
    "where|"
    "location|"
    "Hilfe|"
    "hilfe|"
    "wie|"
    "Wie"
    ")$"
)

JA_PATTERN = re.compile(
    "^("
    "Okay|"
    "ok|"
    "okay|"
    "OK|"
    "Ok|"
    "ja|"
    "Ja|"
    "Jap|"
    "Jo|"
    "Joa|"
    "Yo|"
    "Yap|"
    "Yes|"
    "Yess|"
    "Yesss|"
    "Jawohl|"
    "jawol|"
    "Jawoll|"
    "Auf jeden Fall|"
    "Auf jeden|"
    "Klar|"
    "Klaro|"
    "Ci|"
    "Cí|"
    "Okay|"
    "Ok|"
    "Oki|"
    "kay|"
    "jes|"
    "jep|"
    "yep|"
    "yop|"
    "yup|"
    "yupp|"
    "Bin dabei|"
    "dabie|"
    "Gefunden|"
    "bin da|"
    "bin hier|"
    "da|"
    "hier|"
    "angekommen|"
    "geschafft|"
    "fertig|"
    "done|"
    "👍|"
    "👌|"
    "🤚|"
    "💪|"
    "of course|"
    "made it|"
    "here|"
    "found it|"
    "bereit|"
    "breit|"
    "bin bereit|"
    "ready|"
    "readi|"
    "redy|"
    "kann losgehen|"
    "kan losgehen|"
    "okey|"
    "oke|"
    "abgemacht|"
    "einverstanden|"
    "ein verstanden|"
    "gut|"
    "gud|"
    "jut|"
    "ordnungsgemäß|"
    "gebongt|"
    "gecheckt|"
    "gescheckt|"
    "gechekt|"
    "ist geritzt|"
    "is geritzt|"
    "all right|"
    "allright|"
    "d'accord|"
    "daccord|"
    "find ich gut|"
    "find ich super|"
    "sicher|"
    "sure|"
    "freilich|"
    "freili|"
    "logo|"
    "logen|"
    "na logo|"
    "natürlich|"
    "türlich|"
    "tuerlich|"
    "natuerlich|"
    "immer|"
    "alle Mal|"
    "allemal|"
    "gewiss|"
    "gewiß|"
    "gewis|"
    "fraglos|"
    "wahrlich|"
    "warlich|"
    "ausreichend|"
    "zweifellos|"
    "allerdings|"
    "mit Sicherheit|"
    "topp|"
    "top|"
    "oki doki|"
    "okidoki|"
    "von mir aus|"
    "meinetwegen|"
    "meinet wegen|"
    "wenns sein muss|"
    "wenn’s sein muss|"
    "wens sein muss|"
    "wen’s sein muss|"
    "vermutlich|"
    "wahrscheinlich|"
    "glaube schon|"
    "glaub schon|"
    "bestimmt|"
    "in der Tat|"
    "positiv|"
    ")$"
)


NEIN_PATTERN = re.compile(
    "^("
    "Nein|"
    "nein|"
    "Nope|"
    "nop|"
    "Nee|"
    "ne|"
    "neee|"
    "nain|"
    "nö|"
    "auf keinen Fall|"
    "auf keinsten|"
    "no|"
    "nada|"
    "nien|"
    "nicht|"
    "never|"
    "👎|"
    "🙅‍♀️|"
    "🙅‍♂️|"
    "nimmermehr|"
    "negativ|"
    "veto|"
    "weto|"
    "keinesfalls|"
    "nie und nimmer|"
    "nieundnimmer|"
    "sicher nicht|"
    "unmöglich|"
    "i wo|"
    "mitnichten|"
    "keineswegs|"
    "gar nicht|"
    ")$"
)



from telegram.ext import MessageFilter
from ctparse import ctparse

class FilterUhrzeit(MessageFilter):
    def __init__(self):
        pass

    def filter(self, message):
        try:
            parse = ctparse(message.text, timeout=1)
            return parse.resolution.isDateTime
        except:
            return False


from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate
import yaml

import time

def read_action_yaml(filename, action_functions={}):
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    actions_dict = {}

    for key, value in yaml_dict.items():
        actions_dict[key] = Action(value, action_functions=action_functions)

    return actions_dict


class Action():
    def __init__(self, actions, action_functions={}):
        self.actions = actions
        self.action_functions = action_functions

    def __call__(self, client, update: WhatsAppUpdate, context):
        for item in self.actions:
            time.sleep(.2)
            if item["type"] == "message":
                client.messages.create(
                    body=item["text"].format(
                        **{"profileName": update.ProfileName, "echo": update.Body, **context}),
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "venue":
                client.messages.create(
                    body=item["title"],
                    persistent_action=['geo:{},{}|{}'.format(
                        item["latitude"], item["longitude"], item["address"])],
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "photo":
                client.messages.create(
                    media_url=item["url"],
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "video":
                client.messages.create(
                    media_url=item["url"],
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "media_group":
                client.messages.create(
                    media_url=item["urls"],
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "audio" or item["type"] == "voice":
                client.messages.create(
                    media_url=[item["url"]],
                    from_=update.To,
                    to=update.From
                )
            elif item["type"] == "poll":
                message = item["question"] + "\n"
                for option in item["options"]:
                    message += option + "\n"
                client.messages.create(
                    body=message,
                    from_=update.To,
                    to=update.From
                )

            elif item["type"] == "function":
                arguments = {i: item[i]
                             for i in item if i != 'type' and i != 'func'}
                self.action_functions[item["func"]](
                    client, update, context, **arguments)

            elif item["type"] == "return":
                return item["state"]



import os
from configparser import ConfigParser
from datetime import datetime

import boto3
import requests
from telegram import CallbackQuery

from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate

config = ConfigParser()
config.read("config.ini")

session = boto3.session.Session()
s3_client = session.client('s3',
                        region_name=config["space"]["region_name"],
                        endpoint_url=config["space"]["endpoint_url"],
                        aws_access_key_id=os.getenv('SPACES_KEY'),
                        aws_secret_access_key=os.getenv('SPACES_SECRET'))


def telegram_write_photo(update, context, bucket, folder):
    user_id = update.effective_user.id
    name = update.effective_user.name

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_" +
                      str(user_id) + '.jpg',
                      Body=update.message.photo[-1].get_file(
                      ).download_as_bytearray(),
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def whatsapp_write_photo(client, update: WhatsAppUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName
    update.MediaUrl0

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_" +
                      str(user_id) + '.jpg',
                      Body=requests.get(update.MediaUrl0,
                                        allow_redirects=True).content,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def telegram_write_message(update, context, bucket, folder):
    user_id = update.effective_user.id
    name = update.effective_user.name
    message = update.effective_message.text

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_" +
                      str(user_id) + '.txt',
                      Body=message,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def whatsapp_write_message(client, update: WhatsAppUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName
    message = update.Body

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_"+str(user_id) + '.txt',
                      Body=message,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def telegram_write_voice(update, context, bucket, folder):
    user_id = update.effective_user.id
    name = update.effective_user.name

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_" +
                      str(user_id) + '.mp3',
                      Body=update.message.voice.get_file().download_as_bytearray(),
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def whatsapp_write_voice(client, update: WhatsAppUpdate, context, bucket, folder):
    user_id = update.WaId
    name = update.ProfileName

    s3_client.put_object(Bucket=bucket,
                      Key=folder + "/" +
                      str(datetime.now())+"_"+str(user_id) + '.mp3',
                      Body=requests.get(update.MediaUrl0,
                                        allow_redirects=True).content,
                      ACL='private',
                      # Metadata={
                      #    'x-amz-meta-my-key': 'your-value'
                      # }
                      )


def telegram_write(update, context, bucket, folder):
    if type(update) == CallbackQuery:
        pass
    elif update.message.voice:
        telegram_write_voice(update, context, bucket, folder)
    elif update.message.text:
        telegram_write_message(update, context, bucket, folder)
    elif update.message.photo:
        telegram_write_photo(update, context, bucket, folder)
    else:
        raise NotImplementedError("This type of update can not be saved")


def whatsapp_write(client, update: WhatsAppUpdate, context, bucket, folder):
    if update.MediaContentType0 and update.MediaContentType0.startswith("audio"):
        whatsapp_write_voice(client, update, context, bucket, folder)
    elif update.MediaContentType0 and update.MediaContentType0.startswith("image"):
        whatsapp_write_photo(client, update, context, bucket, folder)
    elif update.Body != "":
        whatsapp_write_message(client, update, context, bucket, folder)
    else:
        raise NotImplementedError("This type of update can not be saved")


telegram_action_functions = {"write_photo": telegram_write_photo,
                             "write_message": telegram_write_message,
                             "write_voice": telegram_write_voice,
                             "write": telegram_write
                             }

whatsapp_action_functions = {"write_photo": whatsapp_write_photo,
                             "write_message": whatsapp_write_message,
                             "write_voice": whatsapp_write_voice,
                             "write": whatsapp_write
                             }




from collections import defaultdict
from digitalguide.standardinteraktion.Action import readActions
from digitalguide.standardinteraktion.Trigger import readTriggers

from digitalguide.standardinteraktion.wegInteraktion import Weg


class Interaktion:
    def __init__(self, name, typ, states, actions, uebertrag_actions, next_interaktion, action_requirements):
        self.name = name
        self.typ = typ
        self.states = states
        self.actions = actions
        self.uebertrag_actions = uebertrag_actions
        self.next_interaktion = next_interaktion

        self.action_requirements = action_requirements

    @classmethod
    def from_excel_dict(cls, excel_dict_item, messanger="Telegram", asset_url=""):
        actions = defaultdict(list)
        states = defaultdict(list)
        action_requirements = []

        poi_interaction, interaktion_dict = excel_dict_item
        poi_name, interaction_typ = poi_interaction

        print(poi_name)

        if "Next" in interaktion_dict.keys():
            weiter_dict = dict((x, y) for x, y in interaktion_dict["Next"])
            next_interaktion = "{}_{}".format(
                weiter_dict["POI"], weiter_dict["Aktion"])
            action_requirements.append(next_interaktion)

        else:
            next_interaktion = None

        if interaction_typ == "Aktion":
            actions["{}".format(poi_name)] = readActions(
                interaktion_dict["Aktion"], messanger=messanger, asset_url=asset_url)

        elif interaction_typ == "Verzweigung":
            actions["{}_weg".format(poi_name)] = readActions(interaktion_dict["Weg"]
                                                             + [("Return", "{}_WEG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)
            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            for aktion, aktion_list in interaktion_dict.items():
                if aktion.startswith("trigger_"):
                    weiter_dict = dict(
                        (x, y) for x, y in interaktion_dict[aktion.replace("trigger_", "Next_")])
                    option_next_interaktion = "{}_{}".format(
                        weiter_dict["POI"], weiter_dict["Aktion"])
                    action_requirements.append(option_next_interaktion)

                    states["{}_WEG".format(poi_name.upper())] += readTriggers(
                        aktion_list, option_next_interaktion)

            states["{}_WEG".format(poi_name.upper())] += [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]

        elif interaction_typ == "Datenabfrage":
            actions["{}_frage".format(poi_name)] = readActions(interaktion_dict["Frage"]
                                                               + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)
            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            states["{}_FRAGE".format(poi_name.upper())] = readTriggers(interaktion_dict["Typ"], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]

        elif interaction_typ == "Weg":
            return Weg(poi_name, interaktion_dict, messanger=messanger, asset_url=asset_url)

        elif interaction_typ == "Quizfrage":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            actions["{}_aufloesung".format(poi_name)] = readActions(
                interaktion_dict["Aufloesung"], messanger=messanger, asset_url=asset_url)\
                + readActions([("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            for aktion, aktion_list in interaktion_dict.items():
                if aktion.startswith("antwort_"):
                    actions["{}_{}".format(poi_name, aktion)] = readActions(aktion_list, messanger=messanger, asset_url=asset_url) \
                        + readActions(interaktion_dict["Aufloesung"], messanger=messanger, asset_url=asset_url) \
                        + readActions([("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

                elif aktion.startswith("trigger_"):
                    states["{}_FRAGE".format(poi_name.upper())] += readTriggers(
                        aktion_list, "{}_{}".format(poi_name, aktion.replace("trigger_", "antwort_")))

            states["{}_FRAGE".format(poi_name.upper())] += readTriggers([("Liste", "Weiter")], "{}_aufloesung".format(poi_name)) \
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]

            states["{}_AUFLOESUNG".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + readTriggers([("Liste", "zurück")], "{}_frage".format(poi_name)) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]

        elif interaction_typ == "Schätzfrage":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            actions["{}_aufloesung".format(poi_name)] = readActions(interaktion_dict["Aufloesung"], messanger=messanger, asset_url=asset_url) \
                + readActions([("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger)

            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            if interaktion_dict["Typ"][0][0] == "Jahreszahl":
                states["{}_FRAGE".format(poi_name.upper())] = (readTriggers([("Regex", "^(\d{1,4})$")], "{}_aufloesung".format(poi_name))) \
                    + [{"handler": "TypeHandler", "type": "Update",
                        "action": "{}_tipp".format(poi_name)}]

            elif interaktion_dict["Typ"][0][0] == "Prozentzahl":
                states["{}_FRAGE".format(poi_name.upper())] = (readTriggers([("Regex", "(\d{1,2}),? ?(\d{,2})")], "{}_aufloesung".format(poi_name))) \
                    + [{"handler": "TypeHandler", "type": "Update",
                        "action": "{}_tipp".format(poi_name)}]

            elif interaktion_dict["Typ"][0][0] == "Kommazahl":
                states["{}_FRAGE".format(poi_name.upper())] = (readTriggers([("Regex", "(\d{1,2}),? ?(\d{,2})")], "{}_aufloesung".format(poi_name))) \
                    + [{"handler": "TypeHandler", "type": "Update",
                        "action": "{}_tipp".format(poi_name)}]

            elif interaktion_dict["Typ"][0][0] == "Römische Jahreszahl":
                states["{}_FRAGE".format(poi_name.upper())] = (readTriggers([("Regex", "^M{0,3}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$")], "{}_aufloesung".format(poi_name))) \
                    + [{"handler": "TypeHandler", "type": "Update",
                        "action": "{}_tipp".format(poi_name)}]

            else:
                print("Der Schätzfragentyp {} ist nicht bekannt!".format(
                    interaktion_dict["Typ"][0][0]))

            states["{}_AUFLOESUNG".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]

        elif interaction_typ == "Listenfrage":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            answer_id_name_list = []
            for aktion, aktion_list in interaktion_dict.items():
                if aktion.startswith("antwort_"):
                    actions["{}_{}".format(poi_name, aktion)] = readActions([("Formel", "function: check_in_context\nkey: {}\nvalue: {}\ndoppelte_antwort: {}\n".format(poi_name, aktion.replace("antwort_", ""), interaktion_dict["doppelte Antwort"][0][1]))], messanger=messanger, asset_url=asset_url)\
                        + readActions([("Formel", "function: append_to_context\nkey: {}\nvalue: {}".format(poi_name, aktion.replace("antwort_", "")))], messanger=messanger, asset_url=asset_url)\
                        + readActions(aktion_list, messanger=messanger, asset_url=asset_url) \
                        + readActions(interaktion_dict["richtig Antwort"], messanger=messanger, asset_url=asset_url)

                elif aktion.startswith("trigger_"):
                    states["{}_FRAGE".format(poi_name.upper())] += readTriggers(
                        aktion_list, "{}_{}".format(poi_name, aktion.replace("trigger_", "antwort_")))

                elif aktion.startswith("name_"):
                    answer_id_name_list.append(
                        [aktion.replace("name_", ""), aktion_list[0][1]])

            actions["{}_aufloesung".format(poi_name)] = [{"type": "function", "func": "eval_list", "answer_id_name_list": answer_id_name_list, "poi": poi_name, "response_text": interaktion_dict["response_text"][0][1]}]\
                + readActions(interaktion_dict["Aufloesung"], messanger=messanger, asset_url=asset_url) \
                + readActions([("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            actions["{}_falsche_antwort".format(poi_name)] = readActions(
                interaktion_dict["falsch Antwort"], messanger=messanger, asset_url=asset_url)

            states["{}_FRAGE".format(poi_name.upper())] += readTriggers([("Liste", "Weiter")], "{}_aufloesung".format(poi_name))\
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_falsche_antwort".format(poi_name)}]

            states["{}_AUFLOESUNG".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]
        elif interaction_typ == "Beteiligungsfrage":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)
            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            actions["{}_aufloesung".format(poi_name)] = readActions(
                interaktion_dict["Aufloesung"] + [("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            states["{}_FRAGE".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + readTriggers([("Liste", "Nein")], next_interaktion) \
                + readTriggers(interaktion_dict["Typ"], "{}_aufloesung".format(poi_name)) \
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]

            states["{}_AUFLOESUNG".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]

        elif interaction_typ == "GIF Generator":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)

            actions["{}_aufloesung".format(poi_name)] = readActions(
                interaktion_dict["Aufloesung"], messanger=messanger, asset_url=asset_url)

            states["{}_FRAGE".format(poi_name.upper())] = readTriggers([("Foto", "")], "{}_aufloesung".format(poi_name)) \
                + readTriggers([("Liste", "Weiter")], next_interaktion) \
                + readTriggers([("Liste", "Nein")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]

        elif interaction_typ == "Infostrecke":
            actions["{}_info".format(poi_name)] = readActions(
                interaktion_dict["Info"] + [("Return", "{}_INFO".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            states["{}_INFO".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]

        elif interaction_typ == "Assoziationskette":
            actions["{}_frage".format(poi_name)] = readActions(
                interaktion_dict["Frage"] + [("Return", "{}_FRAGE".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)
            actions["{}_loop".format(poi_name)] = readActions(
                interaktion_dict["Loop"], messanger=messanger, asset_url=asset_url)
            states["{}_FRAGE".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], "{}_aufloesung".format(poi_name)) \
                + readTriggers([("Freitext", "")], "{}_loop".format(poi_name)) \
                + [{"handler": "TypeHandler", "type": "Update",
                    "action": "{}_tipp".format(poi_name)}]
            actions["{}_tipp".format(poi_name)] = readActions(
                interaktion_dict["Tipp"], messanger=messanger, asset_url=asset_url)
            actions["{}_aufloesung".format(poi_name)] = readActions(
                interaktion_dict["Aufloesung"] + [("Return", "{}_AUFLOESUNG".format(poi_name.upper()))], messanger=messanger, asset_url=asset_url)

            states["{}_AUFLOESUNG".format(poi_name.upper())] = readTriggers([("Liste", "Weiter")], next_interaktion) \
                + [{"handler": "TypeHandler", "type": "Update", "action": "weiter_tipp"}]

        else:
            print("Der Interaktionstyp {} ist nicht bekannt!".format(interaction_typ))

        return cls(poi_name, interaction_typ, states, actions, [], next_interaktion, action_requirements)



from telegram import (Update)
from telegram.ext import (CallbackContext)

from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate


def telegram_default_name(update: Update, context: CallbackContext):
    context.user_data["name"] = update.message.from_user.first_name


def whatsapp_default_name(client, update: WhatsAppUpdate, context):
    context["name"] = update.ProfileName


def telegram_save_text_to_context(update: Update, context: CallbackContext, key):
    context.user_data[key] = update.message.text


def whatsapp_save_text_to_context(client, update: WhatsAppUpdate, context, key):
    context[key] = update.Body


def telegram_save_value_to_context(update: Update, context: CallbackContext, key, value):
    context.user_data[key] = value


def whatsapp_save_value_to_context(client, update: WhatsAppUpdate, context, key, value):
    context[key] = value


telegram_action_functions = {"default_name": telegram_default_name,
                             "save_text_to_context": telegram_save_text_to_context,
                             "save_value_to_context": telegram_save_value_to_context
                             }

whatsapp_action_functions = {"default_name": whatsapp_default_name,
                             "save_text_to_context": whatsapp_save_text_to_context,
                             "save_value_to_context": whatsapp_save_value_to_context
                             }



class WhatsAppUpdate:

    #CombinedMultiDict([ImmutableMultiDict([]), ImmutableMultiDict([('SmsMessageSid', 'SM1cea0a82154a503875f26e1898d972f2'), ('NumMedia', '0'), ('ProfileName', 'Sören'), ('SmsSid', 'SM1cea0a82154a503875f26e1898d972f2'), ('WaId', '4917652163847'), ('SmsStatus', 'received'), ('Body', 'Hi'), ('To', 'whatsapp:+14155238886'), ('NumSegments', '1'), ('MessageSid', 'SM1cea0a82154a503875f26e1898d972f2'), ('AccountSid', 'AC8f3944b5604c359ee4e1c882e64765e2'), ('From', 'whatsapp:+4917652163847'), ('ApiVersion', '2010-04-01')])])
    #CombinedMultiDict([ImmutableMultiDict([]), ImmutableMultiDict([('MediaContentType0', 'image/jpeg'), ('SmsMessageSid', 'MM94942a74f3cfc4fceed98eb6ec1ab15b'), ('NumMedia', '1'), ('ProfileName', 'Sören'), ('SmsSid', 'MM94942a74f3cfc4fceed98eb6ec1ab15b'), ('WaId', '4917652163847'), ('SmsStatus', 'received'), ('Body', ''), ('To', 'whatsapp:+14155238886'), ('NumSegments', '1'), ('MessageSid', 'MM94942a74f3cfc4fceed98eb6ec1ab15b'), ('AccountSid', 'AC8f3944b5604c359ee4e1c882e64765e2'), ('From', 'whatsapp:+4917652163847'), ('MediaUrl0', 'https://api.twilio.com/2010-04-01/Accounts/AC8f3944b5604c359ee4e1c882e64765e2/Messages/MM94942a74f3cfc4fceed98eb6ec1ab15b/Media/ME1eb56f5a20a667e7f4dc8428c04c1c36'), ('ApiVersion', '2010-04-01')])])
    # CombinedMultiDict([ImmutableMultiDict([]), ImmutableMultiDict([('Latitude', '52.406563'), ('Longitude', '12.969897'), ('SmsMessageSid', 'SM69ee3cde6532bdcedc2c44b83ee53f41'), ('NumMedia', '0'), ('ProfileName', 'Sören'), ('SmsSid', 'SM69ee3cde6532bdcedc2c44b83ee53f41'), ('WaId', '4917652163847'), ('SmsStatus', 'received'), ('Body', ''), ('To', 'whatsapp:+14155238886'), ('NumSegments', '1'), ('MessageSid', 'SM69ee3cde6532bdcedc2c44b83ee53f41'), ('AccountSid', 'AC8f3944b5604c359ee4e1c882e64765e2'), ('From', 'whatsapp:+4917652163847'), ('ApiVersio
    
    #CombinedMultiDict([ImmutableMultiDict([]), ImmutableMultiDict([('SmsMessageSid', 'SMfcee22301f5b6f136bb804f55e3b4c0a'), ('NumMedia', '0'), ('ProfileName', 'Sören'), ('SmsSid', 'SMfcee22301f5b6f136bb804f55e3b4c0a'), ('WaId', '4917652163847'), ('SmsStatus', 'received'), ('Body', 'start'), ('To', 'whatsapp:+18066055843'), ('MessagingServiceSid', 'MGd2b35b3b63c7c015da9063754f6087bf'), ('NumSegments', '1'), ('MessageSid', 'SMfcee22301f5b6f136bb804f55e3b4c0a'), ('AccountSid', 'AC8f3944b5604c359ee4e1c882e64765e2'), ('From', 'whatsapp:+4917652163847'), ('ApiVersion', '2010-04-01')])])
    def __init__(self, SmsMessageSid, NumMedia, ProfileName, SmsSid, WaId, SmsStatus, Body, To, NumSegments, MessageSid, AccountSid, From, ApiVersion, MessagingServiceSid=None, MediaContentType0=None, MediaUrl0=None, Latitude=None, Longitude=None) -> None:
        self.Latitude = Latitude
        self.Longitude = Longitude
        self.SmsMessageSid = SmsMessageSid
        self.NumMedia = NumMedia
        self.ProfileName = ProfileName
        self.SmsSid = SmsSid
        self.WaId = WaId
        self.SmsStatus = SmsStatus
        self.Body = Body
        self.To = To
        self.NumSegments = NumSegments
        self.MessageSid = MessageSid
        self.AccountSid = AccountSid
        self.From = From
        self.ApiVersion = ApiVersion
        self.MediaContentType0 = MediaContentType0
        self.MediaUrl0 = MediaUrl0
        self.MessagingServiceSid = MessagingServiceSid



from telegram import Update
from telegram.ext import CallbackContext

import logging
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
    
def error_handler(update: Update, context: CallbackContext):
    """Log the error to notify the developer."""
    logger.error(msg="Exception while handling an update:",
                 exc_info=context.error)


import yaml
from telegram import Update
from telegram.ext import (CommandHandler, Filters, MessageHandler,
                          PollAnswerHandler, TypeHandler)

from digitalguide import rasa_filter, uhrzeit_filter
from digitalguide.pattern import (EMOJI_PATTERN, JA_PATTERN, NEIN_PATTERN,
                                  WEITER_PATTERN, WOHIN_PATTERN)


def read_state_yml(filename, actions=None, prechecks: list = None):
    if actions is None:
        actions = {}
    if prechecks is None:
        prechecks = []
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    states_dict = {}

    for state, handlers in yaml_dict.items():
        handler_list = prechecks[:]
        for handler in handlers:
            if handler["handler"] == "MessageHandler":
                if handler["filter"] == "regex":
                    if handler["regex"] == "EMOJI_PATTERN":
                        newHandler = MessageHandler(Filters.regex(
                            EMOJI_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "WEITER_PATTERN":
                        newHandler = MessageHandler(Filters.regex(
                            WEITER_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "WOHIN_PATTERN":
                        newHandler = MessageHandler(Filters.regex(
                            WOHIN_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "JA_PATTERN":
                        newHandler = MessageHandler(Filters.regex(
                            JA_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "NEIN_PATTERN":
                        newHandler = MessageHandler(Filters.regex(
                            NEIN_PATTERN), actions[handler["action"]])
                    else:
                        newHandler = MessageHandler(Filters.regex(
                            handler["regex"]), actions[handler["action"]])
                elif handler["filter"] == "text":
                    newHandler = MessageHandler(
                        Filters.text, actions[handler["action"]])
                elif handler["filter"] == "photo":
                    newHandler = MessageHandler(
                        Filters.photo, actions[handler["action"]])
                elif handler["filter"] == "voice":
                    newHandler = MessageHandler(
                        Filters.voice, actions[handler["action"]])
                elif handler["filter"] == "rasa":
                    newHandler = MessageHandler(
                        rasa_filter.FilterRasa(handler["intent"]), actions[handler["action"]])
                elif handler["filter"] == "uhrzeit":
                    newHandler = MessageHandler(
                        uhrzeit_filter.FilterUhrzeit(), actions[handler["action"]])
                else:
                    raise NotImplementedError(
                        "This filter is not implemented: {}".format(handler["filter"]))
            elif handler["handler"] == "CommandHandler":
                newHandler = CommandHandler(
                    handler["command"], actions[handler["action"]])
            elif handler["handler"] == "PollAnswerHandler":
                newHandler = PollAnswerHandler(actions[handler["action"]])
            elif handler["handler"] == "TypeHandler":
                if handler["type"] == "Update":
                    type_ = Update
                newHandler = TypeHandler(type_, actions[handler["action"]])
            else:
                raise NotImplementedError(
                    "This Handler is not implemented: {}".format(handler["handler"]))

            handler_list.append(newHandler)

        states_dict[state] = handler_list

    return states_dict



from PIL import Image
from io import BytesIO

from telegram import (Update)
from telegram.ext import (CallbackContext)

from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate


def generate_gif(im1, im2):
    im1 = im1.resize((round(im2.size[0]*1), round(im2.size[1]*1)))
    im1 = im1.convert(im2.mode)

    images = []
    frames = 10

    for i in range(frames+1):
        im = Image.blend(im1, im2, i/frames)
        images.append(im)

    for i in range(frames+1):
        im = Image.blend(im1, im2, 1-i/frames)
        images.append(im)

    bio = BytesIO()
    bio.name = 'image.gif'

    images[0].save(bio, 'GIF', save_all=True,
                   append_images=images[1:], duration=150, loop=0, optimize=True)
    bio.seek(0)
    return bio


def overlay_images(background, foreground):
    foreground = foreground.resize(
        (round(background.size[0]), round(background.size[1])))
    background.paste(foreground, (0, 0), foreground)
    bio = BytesIO()
    bio.name = 'image.png'
    background.save(bio, 'PNG')
    bio.seek(0)
    return bio


def telegram_eval_gif_generation(update: Update, context: CallbackContext, picture):
    im_bytes = update.message.photo[-1].get_file().download_as_bytearray()

    im_file = BytesIO(im_bytes)  # convert image to file-like object
    im1 = Image.open(im_file)   # img is now PIL Image object
    im2 = Image.open('assets/' + picture)

    gif = generate_gif(im1, im2)

    update.message.reply_document(gif)


def whatsapp_eval_gif_generation(client, update: WhatsAppUpdate, context, picture):
    import requests
    import time
    import boto3
    import os

    from configparser import ConfigParser
    config = ConfigParser()
    config.read("config.ini")

    session = boto3.session.Session()
    s3_client = session.client('s3',
                               region_name=config["space"]["region_name"],
                               endpoint_url=config["space"]["endpoint_url"],
                               aws_access_key_id=os.getenv('SPACES_KEY'),
                               aws_secret_access_key=os.getenv('SPACES_SECRET'))

    if update.MediaUrl0:
        im1_bytes = requests.get(
            update.MediaUrl0, allow_redirects=True).content

        im1_file = BytesIO(im1_bytes)  # convert image to file-like object
        im1 = Image.open(im1_file)   # img is now PIL Image object

        im2_bytes = requests.get(config["assets"]["url"] +
                                 "/" + picture).content
        im2_file = BytesIO(im2_bytes)  # convert image to file-like object
        im2 = Image.open(im2_file)

        gif = generate_gif(im2, im1)

        time_str = str(round(time.time() * 1000))

        s3_client.put_object(Bucket=config["assets"]["bucket"],
                             Key="gif" + "/" + time_str + "_" +
                             str(update.WaId) + '.gif',
                             Body=gif,
                             ACL='public-read',
                             ContentType='image/gif'
                             # Metadata={
                             #    'x-amz-meta-my-key': 'your-value'
                             # }
                             )

        client.messages.create(
            media_url=config["assets"]["url"] + "/gif/" +
            time_str + "_" + str(update.WaId) + '.gif',
            from_=update.To,
            to=update.From
        )


telegram_action_functions = {"eval_gif_generation": telegram_eval_gif_generation,
                             }

whatsapp_action_functions = {"eval_gif_generation": whatsapp_eval_gif_generation,
                             }



from google import auth
from telegram.ext import MessageFilter
import requests as req
from requests.exceptions import Timeout
from configparser import ConfigParser 
import os
import json
  
config = ConfigParser() 
config.read('config.ini')

class FilterRasa(MessageFilter):
    def __init__(self, intent, confidence=0.8):
        self.intent = intent
        self.confidence = confidence
        
    def filter(self, message):
        try:
            payload = json.dumps({
            "username": os.getenv('RASA_USER'),
            "password": os.getenv('RASA_PASSWORD')
            })

            token_response = req.post(config["rasa"]["url"] + "/api/auth", data=payload)
            print(token_response.json())

            RASA_TOKEN = token_response.json()["access_token"]
            response = req.post(config["rasa"]["url"] + "/api/projects/default/logs", params={"q": message.text}, headers={'Authorization': 'Bearer {}'.format(RASA_TOKEN)},  timeout = (3, 8))
            print(response.json())
        except (Timeout, KeyError) as e:
            print(e)
            False
        else:
            if not response.ok:
                return False                                                         
            return response.json()["user_input"]["intent"]["name"] == self.intent and response.json()["user_input"]["intent"]["confidence"] >= self.confidence


from setuptools import find_packages, setup

setup(
    name='digitalguide',
    packages=find_packages(),
    version='0.0.187',
    description='A Python Library to write digital guides for telegram',
    author='Soeren Etler',
    license='MIT',
    install_requires=["python-telegram-bot",
                      "ctparse",
                      "boto3",
                      "pymongo[srv]>= 3.4, < 4.0",
                      "mongoengine",
                      "Pillow",
                      "openpyxl",
                      "PyYAML",
                      "twilio",
                      "flask",
                      "requests",
                      "gunicorn",
                      "google-api-python-client",
                      "google-auth-httplib2",
                      "google-auth-oauthlib"],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
)



import json
import os
import re
from configparser import ConfigParser
from typing import List

import requests as req
import yaml
from digitalguide.pattern import (EMOJI_PATTERN, JA_PATTERN, NEIN_PATTERN,
                                  WEITER_PATTERN, WOHIN_PATTERN)
from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate
from requests.exceptions import Timeout


def read_state_yml(filename, actions={}, prechecks: List = []):
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    states_dict = {}

    for state, handlers in yaml_dict.items():
        handler_list = prechecks[:]
        for handler in handlers:
            if handler["handler"] == "PollAnswerHandler":
                newHandler = MessageHandler(RegexFilter(
                    "^(.)+"), actions[handler["action"]])
            elif handler["handler"] == "MessageHandler":
                if handler["filter"] == "regex":
                    if handler["regex"] == "EMOJI_PATTERN":
                        newHandler = MessageHandler(RegexFilter(
                            EMOJI_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "WEITER_PATTERN":
                        newHandler = MessageHandler(RegexFilter(
                            WEITER_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "WOHIN_PATTERN":
                        newHandler = MessageHandler(RegexFilter(
                            WOHIN_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "JA_PATTERN":
                        newHandler = MessageHandler(RegexFilter(
                            JA_PATTERN), actions[handler["action"]])
                    elif handler["regex"] == "NEIN_PATTERN":
                        newHandler = MessageHandler(RegexFilter(
                            NEIN_PATTERN), actions[handler["action"]])
                    else:
                        newHandler = MessageHandler(RegexFilter(
                            handler["regex"]), actions[handler["action"]])
                elif handler["filter"] == "text":
                    newHandler = MessageHandler(
                        TextFilter(), actions[handler["action"]])
                elif handler["filter"] == "photo":
                    newHandler = MessageHandler(
                        PhotoFilter(), actions[handler["action"]])
                elif handler["filter"] == "voice":
                    newHandler = MessageHandler(
                        VoiceFilter(), actions[handler["action"]])
                elif handler["filter"] == "rasa":
                    newHandler = MessageHandler(FilterRasa(
                        handler["intent"]), actions[handler["action"]])
            elif handler["handler"] == "CommandHandler":
                newHandler = CommandHandler(
                    handler["command"], actions[handler["action"]])
            elif handler["handler"] == "TypeHandler":
                if handler["type"] == "Update":
                    type_ = WhatsAppUpdate
                newHandler = TypeHandler(type_, actions[handler["action"]])

            handler_list.append(newHandler)

        states_dict[state] = handler_list

    return states_dict


class MessageHandler:
    def __init__(self, filters, callback):
        self.filters = filters
        self.callback = callback

    def check_update(self, update: WhatsAppUpdate):
        return self.filters(update)


class CommandHandler:
    def __init__(self, command, callback):
        self.command = command
        self.callback = callback

    def check_update(self, update: WhatsAppUpdate):
        if update.Body.strip("/").lower() == self.command.lower():
            return True


class TypeHandler:
    def __init__(self, type_, callback):
        self.type_ = type_
        self.callback = callback

    def check_update(self, update: WhatsAppUpdate):
        return type(update) is self.type_


class RegexFilter:
    def __init__(self, regex) -> None:
        self.regex = regex

    def __call__(self, update: WhatsAppUpdate) -> bool:
        return re.match(self.regex, update.Body)


class TextFilter:
    def __call__(self, update: WhatsAppUpdate) -> bool:
        return update.Body != ""


class PhotoFilter:
    def __call__(self, update: WhatsAppUpdate) -> bool:
        return update.MediaContentType0 and update.MediaContentType0.startswith("image")


class VoiceFilter:
    def __call__(self, update: WhatsAppUpdate) -> bool:
        return update.MediaContentType0 and update.MediaContentType0.startswith("audio")


config = ConfigParser()
config.read('config.ini')


class FilterRasa:
    def __init__(self, intent, confidence=0.8):
        self.intent = intent
        self.confidence = confidence

    def __call__(self, update: WhatsAppUpdate):
        try:
            payload = json.dumps({
                "username": os.getenv('RASA_USER'),
                "password": os.getenv('RASA_PASSWORD')
            })

            token_response = req.post(
                config["rasa"]["url"] + "/api/auth", data=payload)
            print(token_response.json())

            RASA_TOKEN = token_response.json()["access_token"]
            response = req.post(config["rasa"]["url"] + "/api/projects/default/logs", params={
                                "q": update.Body}, headers={'Authorization': 'Bearer {}'.format(RASA_TOKEN)},  timeout=(3, 8))
            print(response.json())
        except (Timeout, KeyError) as e:
            print(e)
            False
        else:
            if not response.ok:
                return False
            return response.json()["user_input"]["intent"]["name"] == self.intent and response.json()["user_input"]["intent"]["confidence"] >= self.confidence



import base64
import logging
import os
import re
from configparser import ConfigParser
from io import BytesIO

import yaml
from telegram import (CallbackQuery, InlineKeyboardButton,
                      InlineKeyboardMarkup, InputFile, InputMediaPhoto,
                      KeyboardButton, ParseMode, Poll, ReplyKeyboardMarkup,
                      ReplyKeyboardRemove, Update)
from telegram.ext import CallbackContext, ConversationHandler

config = ConfigParser()
config.read('config.ini')

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                    level=logging.INFO)

logger = logging.getLogger(__name__)


class callback_query_handler():
    def __init__(self, actions_dict):
        self.actions_dict = actions_dict

    def __call__(self, update: Update, context: CallbackContext):
        query = update.callback_query

        query.answer()
        query.edit_message_reply_markup(InlineKeyboardMarkup([]))
        if query.data.split(":")[0] == "action":
            return self.actions_dict[query.data.split(":")[1]](query, context)
        raise ValueError(
            "This type of operation ({}) is not supported in a callback query handler".format(query.data))


def read_action_yaml(filename, action_functions={}, log_user=True, log_interaction=True):
    with open(filename) as file:
        yaml_dict = yaml.load(file, Loader=yaml.FullLoader)

    actions_dict = {}

    for key, value in yaml_dict.items():
        actions_dict[key] = Action(
            value, action_functions=action_functions, log_user=log_user, log_interaction=log_interaction)

    return actions_dict


class Action():
    def __init__(self, actions, action_functions={}, log_user=True, log_interaction=True):
        self.actions = actions
        self.action_functions = action_functions
        self.log_user = log_user
        self.log_interaction = log_interaction
        if log_user or log_interaction:
            import mongoengine
            dbname = config["bot"]["bot_name"]
            with open("ca-certificate.crt", "w") as text_file:
                text_file.write(os.getenv('DATABASE_CERT'))
            mongoengine.connect(alias=dbname, host="mongodb+srv://" + os.getenv("DATABASE_USERNAME")+":" + os.getenv("DATABASE_PASSWORD") +
                                "@" + os.getenv("DATABASE_HOST") + "/"+dbname+"?authSource=admin&tls=true&tlsCAFile=ca-certificate.crt")

    def __call__(self, update: Update, context: CallbackContext):
        if self.log_user and not ("user_id" in context.user_data.keys()):
            from digitalguide.db_objects import User
            db_user = User(user_id=str(update.effective_user.id),
                           first_name=update.effective_user.first_name,
                           last_name=update.effective_user.last_name,
                           username=update.effective_user.username,
                           language_code=update.effective_user.language_code)
            db_user.save()
            context.user_data["user_id"] = db_user

        if self.log_interaction:
            from digitalguide.db_objects import Interaction
            if type(update) == CallbackQuery:
                if update.message.text:
                    interaction_text = update.message.text
                else:
                    interaction_text = None

                Interaction(user=context.user_data["user_id"],
                            update_id=update.id,
                            first_name=update.from_user.first_name,
                            last_name=update.from_user.last_name,
                            username=update.from_user.username,
                            interaction_full=update.to_dict(),
                            interaction_text=interaction_text,
                            interaction_data=update.data,
                            interaction_id=update.id
                            ).save()
            elif update.poll_answer:
                Interaction(user=context.user_data["user_id"],
                            update_id=update.update_id,
                            interaction_full=update.poll_answer.to_dict(),
                            first_name=update.effective_user.first_name,
                            last_name=update.effective_user.last_name,
                            username=update.effective_user.username,
                            interaction_data=str(
                                update.poll_answer.option_ids),
                            interaction_id=update.poll_answer.poll_id
                            ).save()
            else:
                if update.effective_message.text:
                    message_text = update.effective_message.text
                else:
                    message_text = None

                Interaction(user=context.user_data["user_id"],
                            update_id=update.update_id,
                            interaction_full=update.effective_message.to_dict(),
                            first_name=update.effective_user.first_name,
                            last_name=update.effective_user.last_name,
                            username=update.effective_user.username,
                            interaction_text=message_text,
                            date=update.effective_message.date,
                            interaction_id=update.effective_message.message_id
                            ).save()

        for item in self.actions:
            if "InlineKeyboard" in item:
                keyboard = [[]]
                for button in item["InlineKeyboard"]:
                    if "data" in button:
                        callback_data = button["data"]
                    else:
                        callback_data = None

                    if "url" in button:
                        callback_url = button["url"]
                    else:
                        callback_url = None

                    keyboard[0].append(InlineKeyboardButton(
                        button["text"], callback_data=callback_data, url=callback_url))

                reply_markup = InlineKeyboardMarkup(keyboard)

            elif "ReplyKeyboardMarkup" in item:
                keyboard = [[]]
                for button in item["ReplyKeyboardMarkup"]:
                    if "request_location" in button:
                        request_location = button["request_location"]
                    else:
                        request_location = False

                    keyboard[0].append(KeyboardButton(
                        text=button["text"], request_location=request_location))

                reply_markup = ReplyKeyboardMarkup(
                    keyboard, one_time_keyboard=True)
            else:
                reply_markup = ReplyKeyboardRemove()

            if item["type"] == "message":
                parse_mode = None
                if "parse_mode" in item:
                    parse_mode = item["parse_mode"]
                disable_web_page_preview = None
                if "disable_web_page_preview" in item:
                    disable_web_page_preview = item["disable_web_page_preview"]

                text = item["text"]
                
                print("type: ", type(update))
                print("update: ", update)
                print("context: ", str(context.user_data))
                if type(update) == CallbackQuery:
                    update.from_user.send_message(text.format(
                        **context.user_data), reply_markup=reply_markup, parse_mode=parse_mode, disable_web_page_preview=disable_web_page_preview)
                elif update.callback_query:
                    update.callback_query.from_user.send_message(text.format(
                        **context.user_data), reply_markup=reply_markup, parse_mode=parse_mode, disable_web_page_preview=disable_web_page_preview)
                elif update.poll_answer:
                    update.poll_answer.user.send_message(text.format(
                        **context.user_data), reply_markup=reply_markup, parse_mode=parse_mode, disable_web_page_preview=disable_web_page_preview)
                else:
                    update.message.reply_text(text.format(
                        **context.user_data), reply_markup=reply_markup, parse_mode=parse_mode, disable_web_page_preview=disable_web_page_preview)
            elif item["type"] == "photo":
                if "caption" in item.keys():
                    caption = item["caption"]
                else:
                    caption = None

                if "file" in item.keys():
                    photo = open(item["file"], 'rb')
                else:
                    photo = item["url"]

                if type(update) != CallbackQuery and update.poll_answer:
                    update.poll_answer.user.send_photo(
                        photo, caption=caption, reply_markup=reply_markup)
                else:
                    update.message.reply_photo(
                        photo, caption=caption, reply_markup=reply_markup)

            elif item["type"] == "document":
                if "caption" in item.keys():
                    caption = item["caption"]
                else:
                    caption = None

                if "file" in item.keys():
                    document = open(item["file"], 'rb')
                else:
                    document = item["url"]

                if type(update) != CallbackQuery and update.poll_answer:
                    update.poll_answer.user.send_document(
                        document, caption=caption, reply_markup=reply_markup)
                else:
                    update.message.reply_document(
                        document, caption=caption, reply_markup=reply_markup)

            elif item["type"] == "audio":
                if "file" in item.keys():
                    audio = open(item["file"], 'rb')
                else:
                    audio = item["url"]

                update.message.reply_audio(audio, title=item["title"], performer=item["performer"], reply_markup=reply_markup)

            elif item["type"] == "voice":
                if "caption" in item.keys():
                    caption = item["caption"]
                else:
                    caption = None

                if "file" in item.keys():
                    voice = open(item["file"], 'rb')
                else:
                    voice = item["url"]

                update.message.reply_voice(voice, caption=caption, reply_markup=reply_markup)


            elif item["type"] == "contact":
                update.message.reply_contact(phone_number=item["phone_number"],
                                             first_name=item["first_name"],
                                             last_name=item["last_name"],
                                             reply_markup=reply_markup)
            elif item["type"] == "poll":
                if "poll_type" in item.keys():
                    poll_type = item["poll_type"]
                else:
                    poll_type = "quiz"
                if "correct_option_id" in item.keys():
                    correct_option_id = item["correct_option_id"]
                else:
                    correct_option_id = None
                update.message.reply_poll(question=item["question"],
                                          options=item["options"],
                                          type=poll_type,
                                          correct_option_id=correct_option_id,
                                          is_anonymous=False
                                          )

            elif item["type"] == "media_group":
                if "files" in item.keys():
                    photoGroup = [InputMediaPhoto(media=open(
                        photo, 'rb')) for photo in item["files"]]
                else:
                    photoGroup = [InputMediaPhoto(media=open(
                        photo, 'rb')) for photo in item["urls"]]
                update.message.reply_media_group(media=photoGroup)
            elif item["type"] == "sticker":
                if type(update) != CallbackQuery and update.poll_answer:
                    update.poll_answer.user.send_sticker(item["id"])
                else:
                    update.message.reply_sticker(item["id"])
            elif item["type"] == "video":
                if "caption" in item.keys():
                    caption = item["caption"]
                else:
                    caption=None
                if "file" in item.keys():
                    video = open(item["file"], 'rb')
                else:
                    video = item["url"]
                update.message.reply_video(video,caption=caption)

            ## type: venue
            # latitude: 52.4090401
            # longitude: 12.9724552
            # address: Bahnhof Golm
            # title: Start der Rallye

            elif item["type"] == "venue":
                if type(update) != CallbackQuery and update.poll_answer:
                    update.poll_answer.user.send_venue(
                        latitude=item["latitude"], longitude=item["longitude"], address=item["address"], title=item["title"])
                else:
                    update.message.reply_venue(
                        latitude=item["latitude"], longitude=item["longitude"], address=item["address"], title=item["title"])

            elif item["type"] == "return":
                if item["state"] == "END":
                    return ConversationHandler.END
                return item["state"]
            elif item["type"] == "callback":
                query = update.callback_query

                query.answer()
                query.edit_message_reply_markup(InlineKeyboardMarkup([]))
                for case in item["conditions"]:
                    if query.data == case["condition"]:
                        return Action(case["action"])(query, context)
            elif item["type"] == "function":
                arguments = {i: item[i]
                             for i in item if i != 'type' and i != 'func'}
                result_value = self.action_functions[item["func"]](
                    update, context, **arguments)
                if result_value:
                    return result_value



from telegram import (Update, ReplyKeyboardRemove)
from telegram.ext import (CallbackContext)

from digitalguide.whatsapp.WhatsAppUpdate import WhatsAppUpdate

def telegram_eval_jahreszahl(update: Update, context: CallbackContext, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    import re
    echter_wert = int(echter_wert)
    schaetzung = int(re.findall(r"\d{1,4}", update.message.text)[0])
    if schaetzung == echter_wert:
        update.message.reply_text(richtig_text,
                                  reply_markup=ReplyKeyboardRemove())

    differenz = schaetzung - echter_wert
    if differenz == -1:
        update.message.reply_text(spaeter_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz < -1:
        update.message.reply_text(spaeter_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz == 1:
        update.message.reply_text(vorher_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz > 1:
        update.message.reply_text(vorher_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())


def whatsapp_eval_jahreszahl(client, update: WhatsAppUpdate, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    import re
    echter_wert = int(echter_wert)
    schaetzung = int(re.findall(r"\d{1,4}", update.Body)[0])
    if schaetzung == echter_wert:
        client.messages.create(
            body=richtig_text,
            from_=update.To,
            to=update.From
        )

    differenz = schaetzung - echter_wert
    if differenz == -1:
        client.messages.create(
            body=spaeter_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz < -1:
        client.messages.create(
            body=spaeter_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )

    elif differenz == 1:
        client.messages.create(
            body=vorher_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz > 1:
        client.messages.create(
            body=vorher_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )


def telegram_eval_prozentzahl(update: Update, context: CallbackContext, echter_wert, richtig_text, falsch_text):
    echter_wert = float(echter_wert)
    import re
    match = re.search(
        r"(?P<vorkomma>\d{1,2}),? ?(?P<nachkomma>\d{,2})", update.message.text)

    schaetzung = int(match.group('vorkomma')) + \
        float("0."+match.group('nachkomma'))

    if schaetzung == echter_wert:
        update.message.reply_text(richtig_text,
                                  reply_markup=ReplyKeyboardRemove())
    else:
        update.message.reply_text(falsch_text,
                                  reply_markup=ReplyKeyboardRemove())

def whatsapp_eval_prozentzahl(client, update: WhatsAppUpdate, context, echter_wert, richtig_text, falsch_text):
    echter_wert = float(echter_wert)
    import re
    match = re.search(
        r"(?P<vorkomma>\d{1,2}),? ?(?P<nachkomma>\d{,2})", update.Body)

    schaetzung = int(match.group('vorkomma')) + \
        float("0."+match.group('nachkomma'))

    if schaetzung == echter_wert:
        client.messages.create(
            body=richtig_text,
            from_=update.To,
            to=update.From
        )

    else:
        client.messages.create(
            body=falsch_text,
            from_=update.To,
            to=update.From
        )

def telegram_eval_kommazahl(update: Update, context: CallbackContext, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = float(echter_wert)
    import re
    match = re.search(
        r"(?P<vorkomma>\d{1,2}),? ?(?P<nachkomma>\d{,2})", update.message.text)

    schaetzung = int(match.group('vorkomma')) + \
        float("0."+match.group('nachkomma'))

    if schaetzung == echter_wert:
        update.message.reply_text(richtig_text,
                                  reply_markup=ReplyKeyboardRemove())
    
    differenz = schaetzung - echter_wert
    if differenz == -1:
        update.message.reply_text(spaeter_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz < -1:
        update.message.reply_text(spaeter_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz == 1:
        update.message.reply_text(vorher_singular_text,
                                  reply_markup=ReplyKeyboardRemove())
    elif differenz > 1:
        update.message.reply_text(vorher_plural_text.format(abs(differenz)),
                                  reply_markup=ReplyKeyboardRemove())

def whatsapp_eval_kommazahl(client, update: WhatsAppUpdate, context, echter_wert, richtig_text, vorher_singular_text, vorher_plural_text, spaeter_singular_text, spaeter_plural_text):
    echter_wert = float(echter_wert)
    import re
    match = re.search(
        r"(?P<vorkomma>\d{1,2}),? ?(?P<nachkomma>\d{,2})", update.Body)

    schaetzung = int(match.group('vorkomma')) + \
        float("0."+match.group('nachkomma'))

    if schaetzung == echter_wert:
        client.messages.create(
            body=richtig_text,
            from_=update.To,
            to=update.From
        )

    differenz = schaetzung - echter_wert
    if differenz == -1:
        client.messages.create(
            body=spaeter_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz < -1:
        client.messages.create(
            body=spaeter_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )

    elif differenz == 1:
        client.messages.create(
            body=vorher_singular_text,
            from_=update.To,
            to=update.From
        )

    elif differenz > 1:
        client.messages.create(
            body=vorher_plural_text.format(abs(differenz)),
            from_=update.To,
            to=update.From
        )


telegram_action_functions = {"eval_jahreszahl": telegram_eval_jahreszahl,
                             "eval_prozentzahl": telegram_eval_prozentzahl,
                             "eval_kommazahl": telegram_eval_kommazahl,
                             }

whatsapp_action_functions = {"eval_jahreszahl": whatsapp_eval_jahreszahl,
                             "eval_prozentzahl": whatsapp_eval_prozentzahl,
                             "eval_kommazahl": whatsapp_eval_kommazahl,
                             }



import pytest
from digitalguide.uhrzeit_filter import FilterUhrzeit
from telegram import Message
from telegram import Chat
from datetime import datetime

@pytest.fixture
def uhrzeit_filter():
    '''Filter for the identification of datetimes'''
    return FilterUhrzeit()

@pytest.mark.parametrize("message_text", [
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="0 Uhr")),
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="00:00")),
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="12"))
])
def test_uhrzeit_filter_on_messages(uhrzeit_filter, message_text):
    assert uhrzeit_filter.filter(message_text) is True

@pytest.mark.parametrize("message_text", [
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="weiß nicht")),
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="ohne")),
    (Message(121123, date=datetime.now, chat=Chat(id=1212, type ="private"), text ="leer"))
])
def test_uhrzeit_filter_on_false_messages(uhrzeit_filter, message_text):
    assert uhrzeit_filter.filter(message_text) is False


from telegram import (ReplyKeyboardRemove,InlineKeyboardButton, InlineKeyboardMarkup, Update)
from telegram.ext import (ConversationHandler, CallbackContext)

from digitalguide.contextActions import telegram_default_name, telegram_save_value_to_context

def telegram_return_end(update: Update, context: CallbackContext):
    return ConversationHandler.END

def telegram_entry_conversation(update: Update, context: CallbackContext):
    if context.args:
        telegram_default_name(update,context)
        telegram_save_value_to_context(update,context, "data", True)
        keyboard = [[InlineKeyboardButton(
            "🐾 Los", callback_data='action:' + context.args[0])]]

        reply_markup = InlineKeyboardMarkup(keyboard)
        update.message.reply_text(
            'Hi, freut mich, dass du dabei bist! Mit einem klick auf "los" kannst du direkt an die richtige Position in der Route springen',
            reply_markup=reply_markup)
        return "NONE"

telegram_action_functions = {"return_end": telegram_return_end,
                    "entry_conversation": telegram_entry_conversation
                    }


from telegram.ext import BasePersistence
from collections import defaultdict
from copy import deepcopy
from telegram.utils.helpers import decode_user_chat_data_from_json, decode_conversations_from_json, encode_conversations_to_json
import mongoengine
import json
from bson import json_util
import os

class Conversations(mongoengine.Document):
    obj = mongoengine.DictField()
    meta = { 'collection': 'Conversations', 'ordering': ['-id']}

class UserData(mongoengine.Document):
    obj = mongoengine.DictField()
    meta = { 'collection': 'UserData', 'ordering': ['-id']}

class ChatData(mongoengine.Document):
    obj = mongoengine.DictField()
    meta = { 'collection': 'ChatData', 'ordering': ['-id']}

class BotData(mongoengine.Document):
    obj = mongoengine.DictField()
    meta = { 'collection': 'BotData', 'ordering': ['-id']}

from pymongo import monitoring

class CommandLogger(monitoring.CommandListener):

    def started(self, event):
        print("Command {0.command_name} with request id "
                 "{0.request_id} started on server "
                 "{0.connection_id}".format(event))

    def succeeded(self, event):
        print("Command {0.command_name} with request id "
                 "{0.request_id} on server {0.connection_id} "
                 "succeeded in {0.duration_micros} "
                 "microseconds".format(event))

    def failed(self, event):
        print("Command {0.command_name} with request id "
                 "{0.request_id} on server {0.connection_id} "
                 "failed in {0.duration_micros} "
                 "microseconds".format(event))

monitoring.register(CommandLogger())

class DBHelper():
    """Class to add and get documents from a mongo database using mongoengine
    """
    def __init__(self, dbname):
        with open("ca-certificate.crt", "w") as text_file:
            text_file.write(os.getenv('DATABASE_CERT'))
        mongoengine.connect(host="mongodb+srv://"+ os.getenv("DATABASE_USERNAME")+":"+ os.getenv("DATABASE_PASSWORD") + "@" +os.getenv("DATABASE_HOST") +"/"+dbname+"?authSource=admin&tls=true&tlsCAFile=ca-certificate.crt")
    
    def add_item(self, data, collection):
        if collection == "Conversations":
            document = Conversations(obj=data)
        elif collection == "UserData":
            document = UserData(obj=data)
        elif collection == "ChatData":
            document = ChatData(obj=data)
        else:
            document = BotData(obj=data)
        document.save()
    def get_item(self, collection):
        if collection == "Conversations":
            document = Conversations.objects()
        elif collection == "UserData":
            document = UserData.objects()
        elif collection == "ChatData":
            document = ChatData.objects()
        else:
            document = BotData.objects()
        if document.first() is None:
            document = {}
        else:
            document = document.first()['obj']

        return document
    def close(self):
        mongoengine.disconnect()

class DBPersistence(BasePersistence):
    """Uses DBHelper to make the bot persistant on a database.
       It's heavily inspired on PicklePersistence from python-telegram-bot
    """
    def __init__(self, dbname):
        super(DBPersistence, self).__init__(store_user_data=True,
                                               store_chat_data=True,
                                               store_bot_data=True)
        self.persistdb = dbname
        self.conversation_collection = "Conversations"
        self.user_data_collection = "UserData"
        self.chat_data_collection = "ChatData"
        self.bot_data_collection = "BotData"
        self.db = DBHelper(dbname=dbname)
        self.user_data = None
        self.chat_data = None
        self.bot_data = None
        self.conversations = None
        self.on_flush = False

    def get_conversations(self, name):
        if self.conversations:
            pass
        else:
            conversations_json = json_util.dumps(self.db.get_item(self.conversation_collection))
            self.conversations = decode_conversations_from_json(conversations_json)
        return self.conversations.get(name, {}).copy()

    def update_conversation(self, name, key, new_state):
        if self.conversations.setdefault(name, {}).get(key) == new_state:
            return
        self.conversations[name][key] = new_state
        if not self.on_flush:
            conversations_json = json_util.loads(encode_conversations_to_json(self.conversations))
            self.db.add_item(conversations_json, self.conversation_collection)

    def get_user_data(self):
        if self.user_data:
            pass
        else:
            user_data_json = json_util.dumps(self.db.get_item(self.user_data_collection))
            if user_data_json != '{}':
                self.user_data = decode_user_chat_data_from_json(user_data_json)
            else:
                self.user_data = defaultdict(dict,{})
        return deepcopy(self.user_data)

    def update_user_data(self, user_id, data):
        if self.user_data is None:
            self.user_data = defaultdict(dict)
        # comment next line if you want to save to db every time this function is called
        if self.user_data.get(user_id) == data:
            return
        self.user_data[user_id] = data
        if not self.on_flush:
            user_data_json = json_util.loads(json.dumps(self.user_data))
            self.db.add_item(user_data_json, self.user_data_collection)

    def get_chat_data(self):
        if self.chat_data:
            pass
        else:
            chat_data_json = json_util.dumps(self.db.get_item(self.chat_data_collection))
            if chat_data_json != "{}":
                self.chat_data = decode_user_chat_data_from_json(chat_data_json)
            else:
                self.chat_data = defaultdict(dict,{})
        return deepcopy(self.chat_data)

    def update_chat_data(self, chat_id, data):
        if self.chat_data is None:
            self.chat_data = defaultdict(dict)
        # comment next line if you want to save to db every time this function is called
        if self.chat_data.get(chat_id) == data:
            return
        self.chat_data[chat_id] = data
        if not self.on_flush:
            chat_data_json = json_util.loads(json.dumps(self.chat_data))
            self.db.add_item(chat_data_json, self.chat_data_collection)

    def get_bot_data(self):
        if self.bot_data:
            pass
        else:
            bot_data_json = json_util.dumps(self.db.get_item(self.bot_data_collection))
            self.bot_data = json.loads(bot_data_json)
        return deepcopy(self.bot_data)

    def update_bot_data(self, data):
        if self.bot_data == data:
            return
        self.bot_data = data.copy()
        if not self.on_flush:
            bot_data_json = json_util.loads(json.dumps(self.bot_data))
            self.db.add_item(self.bot_data, self.bot_data_collection)

    def flush(self):
        if self.conversations:
            conversations_json = json_util.loads(encode_conversations_to_json(self.conversations))
            self.db.add_item(conversations_json, self.conversation_collection)
        if self.user_data:
            user_data_json = json_util.loads(json.dumps(self.user_data))
            self.db.add_item(user_data_json, self.user_data_collection)
        if self.chat_data:
            chat_data_json = json_util.loads(json.dumps(self.chat_data))
            self.db.add_item(chat_data_json, self.chat_data_collection)
        if self.bot_data:
            bot_data_json = json_util.loads(json.dumps(self.bot_data))
            self.db.add_item(self.bot_data, self.bot_data_collection)
        self.db.close()